/*
Halo Query v1.01
Copyright (c) 2003, Brian P. Hurley, brian@fieldeffect.net
10/6/2003 - Initial Release
10/6/2003 - Modified by Andrew Chen, Divo Networks - http://www.divo.net/
			Added Linux Support.  Compile with:
			gcc -o HaloQuery -O4 -DLINUX HaloQuery.c
10/7/2003 - v1.01
            Modified Linux Support
            Added BSD/Other Unix Support
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer. Redistributions in binary
form must reproduce the above copyright notice, this list of conditions and
the following disclaimer in the documentation and/or other materials
provided with the distribution. None of the contributors may be used to endorse
or  promote products derived from this software without specific prior written
permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/

#ifdef WIN32
#include <winsock.h>
#define ETIMEDOUT WSAETIMEDOUT
#define snprintf _snprintf
#define close(s) closesocket(s)
#define SockLastError() WSAGetLastError()
#else
#include <sys/errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#define SOCKET_ERROR -1
#ifndef INADDR_NONE
#define INADDR_NONE ~0
#endif
#define SockLastError() errno
#define SOCKET int
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Configuration Options */
#define VERSION_STRING "v1.01"
#define MAX_DATA_SIZE 2048 /* Max Possible: 65499 bytes */
#define TIMEOUT_VALUE 1 /* In secods */
#define REQUEST_ID 'b','p','h'
#define HALO_QUERY {0xFE,0xFD,0x00,REQUEST_ID,0x13,0xFF,0xFF,0xFF}

typedef struct PlayerInfo {
	char *name;
	char *score;
	char *ping;
	char *team;
} PLAYER_INFO;

typedef struct TeamInfo {
	char *name;
	char *score;
} TEAM_INFO;

typedef struct ServerInfo {
	char addr[65];
	char port[6];
	char *hostname;
	char *gamever;
	char *hostport;
	char *maxplayers;
	char *password;
	char *mapname;
	char *dedicated;
	char *gamemode;
	char *game_classic;
	char *numplayers;
	char *gametype;
	char *teamplay;
	char *gamevariant;
	char *fraglimit;
	char *player_flags;
	char *game_flags;
	PLAYER_INFO player[16];
	TEAM_INFO team[2];
} SERVER_INFO;


unsigned long LookupServer(const char *);
char *nextDetail(char **);
char *nextToken(char **);
void ParseInfo(unsigned char*, int, SERVER_INFO *);
void PrintHeader(SERVER_INFO *, unsigned long);
void PrintBody(SERVER_INFO *);
void PrintFooter();
char *QueryServer(SOCKET, SERVER_INFO *, unsigned char *, int *);



int main() {
    SOCKET fd;
	unsigned char status[MAX_DATA_SIZE];
	int statusSize;
	unsigned long refresh = 0;
	size_t settingSize;
	SERVER_INFO si;
	char *lastError, *queryString, *setting, *pos, strRefresh[11];

#ifdef WIN32
    WSADATA wsaData; 
	if(WSAStartup(MAKEWORD(2, 2), &wsaData) != 0) {
		PrintHeader(&si, refresh);
		printf("ERROR");
		printf("            <span class=\"ltblue\">ERROR:</span> WSAStartup failed to initialize.\n");
		return(1);
	}
#endif

	/* Clear all the fields of Server Information */
	memset(&si, 0, sizeof(si));

	/* Read Input String */
	queryString = getenv("QUERY_STRING");

	*si.addr = 0;
	*si.port = 0;
	if(queryString) {
		/* Read Server Address */
		setting = strstr(queryString, "address=");
		if(setting) {
			setting += 8;
			for(pos = setting; *pos && *pos != '&'; pos++);
			settingSize = pos - setting;
			if(settingSize > sizeof(si.addr)) settingSize = sizeof(si.addr) - 1;
			si.addr[settingSize] = 0;
			strncpy(si.addr, setting, settingSize);
		}
		
		/* Read Server Port */
		setting = strstr(queryString, "port=");
		if(setting) {
			setting += 5;
			for(pos = setting; *pos && *pos != '&'; pos++);
			settingSize = pos - setting;
			if(settingSize > sizeof(si.port)) settingSize = sizeof(si.port) - 1;
			si.port[settingSize] = 0;
			strncpy(si.port, setting, settingSize);
		}

		/* Read Refresh Rate */
		setting = strstr(queryString, "refresh=");
		if(setting) {
			setting += 8;
			for(pos = setting; *pos && *pos != '&'; pos++);
			settingSize = pos - setting;
			if(settingSize > sizeof(strRefresh)) settingSize = sizeof(strRefresh) - 1;
			strRefresh[settingSize] = 0;
			strncpy(strRefresh, setting, settingSize);
			refresh = atol(strRefresh);
		}
	}
	
	/* Fill in Defaults if Needed */
	if(!*si.addr) gethostname(si.addr, sizeof(si.addr));
	if(!*si.port) strcpy(si.port, "2302");

    fd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);

    if(SOCKET_ERROR == fd) {
		PrintHeader(&si, refresh);
		printf("            <span class=\"ltblue\">ERROR:</span> The socket was unable to be created.\n");
        PrintFooter();
		return(1);
	}

	statusSize = sizeof(status);
	lastError = QueryServer(fd, &si, status, &statusSize); 

	close(fd);

#ifdef WIN32
	WSACleanup();
#endif

	if(*lastError) {
		PrintHeader(&si, refresh);
		printf("            <span class=\"ltblue\">ERROR:</span> %s\n", lastError);
		PrintFooter();
	}
	else {
		ParseInfo(status, statusSize, &si);
		PrintHeader(&si, refresh);
		PrintBody(&si);
		PrintFooter();
	}
    
    return(0);
}


unsigned long LookupServer(const char *strServerAddr) {
	struct hostent *host = NULL;
	unsigned long serverAddr = inet_addr(strServerAddr);
    
	/* Resolve Address */
	if(INADDR_NONE == serverAddr) {
        host = gethostbyname(strServerAddr);
        if(host == NULL) return(INADDR_NONE);
        serverAddr = *((u_long *)host->h_addr_list[0]);
    }

    return(serverAddr);
}


char *nextDetail(char **src) {
	nextToken(src);
	return(nextToken(src));
}


char *nextToken(char **src) {
	char *pos = *src;
	char *ret = pos;
	while(*pos) pos++;
	*src = ++pos; /* Skip trailing null character */
	return(ret);
}


/* No error checking is done here.  It is assumed that the data is valid. */
void ParseInfo(unsigned char *status, int statusSize, SERVER_INFO *si) {
	char *currPos, *endPos;
	int nPlayer;

	currPos = (char *)status;
	endPos = currPos + statusSize;

	/* Get Server Information */
	currPos += 5; /* Skip ID */
	si->hostname = nextDetail(&currPos);
	si->gamever = nextDetail(&currPos);
	si->hostport = nextDetail(&currPos);
	si->maxplayers = nextDetail(&currPos);
	si->password = nextDetail(&currPos);
	si->mapname = nextDetail(&currPos);
	si->dedicated = nextDetail(&currPos);
	si->gamemode = nextDetail(&currPos);
	si->game_classic = nextDetail(&currPos);
	si->numplayers = nextDetail(&currPos);
	si->gametype = nextDetail(&currPos);
	si->teamplay = nextDetail(&currPos);
	si->gamevariant = nextDetail(&currPos);
	si->fraglimit = nextDetail(&currPos);
	si->player_flags = nextDetail(&currPos);
	si->game_flags = nextDetail(&currPos);

	/* Get Player Info */
	currPos += 3;
	nextToken(&currPos); /* player_ */
	nextToken(&currPos); /* score_ */
	nextToken(&currPos); /* ping_ */
	nextToken(&currPos); /* team_ */
	currPos++;

	nPlayer = 0;
	while(*currPos) {
		si->player[nPlayer].name = nextToken(&currPos);
		si->player[nPlayer].score = nextToken(&currPos);
		si->player[nPlayer].ping = nextToken(&currPos);
		si->player[nPlayer].team = nextToken(&currPos);
		nPlayer++;
	}

	/* Get Team Info */
	currPos += 2;
	nextToken(&currPos); /* team_t */
	nextToken(&currPos); /* score_t */
	currPos++;

	while(currPos < endPos && *currPos) {
		si->team[0].name = nextToken(&currPos);
		si->team[0].score = nextToken(&currPos);
		si->team[1].name = nextToken(&currPos);
		si->team[1].score = nextToken(&currPos);
	}
}


void PrintHeader(SERVER_INFO *si, unsigned long refresh) {
	printf("Content-type: text/html\n\n");
	printf("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
	printf("<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.1//EN\"\n");
	printf("    \"http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd\">\n");
	printf("<html>\n");
	printf("<head>\n");
	printf("    <meta http-equiv=\"content-type\" content=\"text/html;charset=UTF-8\"/>\n");
	if(refresh) printf("    <meta http-equiv=\"refresh\" content=\"%ld\"/>\n", refresh);
	printf("    <meta name=\"author\" content=\"Brian Hurley\"/>\n");
	printf("    <meta name=\"email\" content=\"brian@fieldeffect.net\"/>\n");
	printf("    <meta name=\"copyright\" content=\"Brian Hurley\"/>\n");
	printf("    <meta name=\"generator\" content=\"Halo Query %s\"/>\n", VERSION_STRING);
	printf("    <meta name=\"language\" content=\"EN-US\"/>\n");
	printf("    <meta name=\"robots\" content=\"nofollow,noindex\"/>\n");
	printf("    <title>Halo Ingame Carnage Report for ");
	printf("%s", si->hostname ? si->hostname : si->addr);
	printf("</title>\n");
	printf("    <style type=\"text/css\">\n");
	printf("        .all {\n");
	printf("            font-family: \"Lucida Console\", Verdana, Sans-Serif;\n");
	printf("            font-size: 16pt;\n");
	printf("            color: white;\n");
	printf("            background-color: black;\n");
	printf("            text-align: left;\n");
	printf("        }\n");
	printf("        .ltblue {\n");
	printf("            color: #66CCFF;\n");
	printf("            font-weight: bold;\n");
	printf("        }\n");
	printf("        .title {\n");
	printf("            font-size: 26pt;\n");
	printf("            font-weight: bold;\n");
	printf("            color: #3399FF;\n");
	printf("            text-align: center;\n");
	printf("        }\n");
	printf("        .infotext {\n");
	printf("            font-size: 8pt;\n");
	printf("            text-align: center;\n");
	printf("        }\n");
	printf("        .red { color: #CC6666; }\n");
	printf("        .blue { color: #6666CC; }\n");
	printf("        .center { text-align: center; }\n");
	printf("        .info { background-color: #000033 }\n");
	printf("        .layouttd { padding-left: 0px; }\n");
	printf("        table { margin-bottom: 16pt; }\n");
	printf("        td { padding-left: 16pt;}\n");
	printf("        hr { color: #006699; }\n");
	printf("        img { border: thin solid #006699; }\n");
	printf("    </style>\n");
	printf("</head>\n");
	printf("<body class=\"all\">\n");
	printf("    <div class=\"title\">HALO INGAME CARNAGE REPORT</div>\n");
	printf("    <div class=\"info\">\n");
	printf("        <hr/>\n");
}


void PrintBody(SERVER_INFO *si) {
	int nPlayers, i, nTeam;
	FILE *imgFile;
	char imgName[32];

	printf("        <table>\n");
	printf("            <tr>\n");
	printf("                <td class=\"layouttd\">\n");
	printf("                    <table>\n");
	printf("                        <tr>\n");
	printf("                            <td class=\"ltblue\">Server Name:</td>\n");
	printf("                            <td>%s</td>\n", si->hostname);
	printf("                        </tr>\n");
	printf("                        <tr>\n");
	printf("                            <td class=\"ltblue\">Server Address:</td>\n");
	printf("                            <td>%s:%s</td>\n", si->addr, si->port);
	printf("                        </tr>\n");
	printf("                        <tr>\n");
	printf("                            <td class=\"ltblue\">Game Version:</td>\n");
	printf("                            <td>%s</td>\n", si->gamever);
	printf("                        </tr>\n");
	printf("                        <tr>\n");
	printf("                            <td class=\"ltblue\">Maximum Players:</td>\n");
	printf("                            <td>%s</td>\n", si->maxplayers);
	printf("                        </tr>\n");
	printf("                        <tr>\n");
	printf("                            <td class=\"ltblue\">Password Protected:</td>\n");
	printf("                            <td>%s</td>\n", ('1' == *si->password) ? "Yes" : "No");
	printf("                        </tr>\n");
	printf("                        <tr>\n");
	printf("                            <td class=\"ltblue\">Current Map:</td>\n");
	printf("                            <td>%s</td>\n", si->mapname);
	printf("                        </tr>\n");
	printf("                        <tr>\n");
	printf("                            <td class=\"ltblue\">Dedicated Server:</td>\n");
	printf("                            <td>%s</td>\n", ('1' == *si->dedicated) ? "Yes" : "No");
	printf("                        </tr>\n");
	printf("                        <tr>\n");
	printf("                            <td class=\"ltblue\">Game Mode:</td>\n");
	printf("                            <td>%s</td>\n", si->gamemode);
	printf("                        </tr>\n");
	printf("                        <tr>\n");
	printf("                            <td class=\"ltblue\">Classic Game:</td>\n");
	printf("                            <td>%s</td>\n", ('1' == *si->game_classic) ? "Yes" : "No");
	printf("                        </tr>\n");
	printf("                        <tr>\n");
	printf("                            <td class=\"ltblue\">Number of Players:</td>\n");
	printf("                            <td>%s</td>\n", si->numplayers);
	printf("                        </tr>\n");
	printf("                        <tr>\n");
	printf("                            <td class=\"ltblue\">Game Type:</td>\n");
	printf("                            <td>%s</td>\n", si->gametype);
	printf("                        </tr>\n");
	printf("                        <tr>\n");
	printf("                            <td class=\"ltblue\">Team Play:</td>\n");
	printf("                            <td>%s</td>\n", ('1' == *si->teamplay) ? "Yes" : "No");
	printf("                        </tr>\n");
	if(strlen(si->gamevariant)) {
		printf("                        <tr>\n");
		printf("                            <td class=\"ltblue\">Game Variant:</td>\n");
		printf("                            <td>%s</td>\n", si->gamevariant);
		printf("                        </tr>\n");
	}
	printf("                        <tr>\n");
	printf("                            <td class=\"ltblue\">Score Limit:</td>\n");
	printf("                            <td>%s</td>\n", si->fraglimit);
	printf("                        </tr>\n");
	printf("                    </table>\n");
	printf("                </td>\n");
	printf("                <td>\n");
	printf("                    <img src=\"");
	snprintf(imgName, sizeof(imgName), "./mapname/%s.gif", si->mapname);
	imgFile = fopen(imgName, "rb");
	if(imgFile == NULL)
		printf("./mapname/Unknown.gif");
	else {
		printf("%s", imgName);
		fclose(imgFile);		
	}
	printf("\"/>\n");
	printf("                    <br/>\n");
	printf("                    <img src=\"");
	snprintf(imgName, sizeof(imgName), "./gametype/%s.gif", si->gametype);
	imgFile = fopen(imgName, "rb");
	if(imgFile == NULL)
		printf("./gametype/Unknown.gif");
	else {
		printf("%s", imgName);
		fclose(imgFile);		
	}
	printf("\"/>\n");
	printf("                </td>\n");
	printf("            </tr>\n");
	printf("        </table>\n");
	if('1' == *si->teamplay) {
		printf("        <table>\n");
		printf("            <tr class=\"ltblue\">\n");
		printf("                <td>Team</td>\n");
		printf("                <td>Score</td>\n");
		printf("            </tr>\n");
		printf("            <tr>\n");
		printf("                <td class=\"red\">Red</td>\n");
		printf("                <td>%s</td>\n", si->team[0].score);
		printf("            </tr>\n");
		printf("            <tr>\n");
		printf("                <td class=\"blue\">Blue</td>\n");
		printf("                <td>%s</td>\n", si->team[1].score);
		printf("            </tr>\n");
		printf("        </table>\n");
	}
	printf("        <table>\n");
	printf("            <tr class=\"ltblue\">\n");
	printf("                <td>Number</td>\n");
	printf("                <td>Name</td>\n");
	printf("                <td>Score</td>\n");
	if('1' == *si->teamplay)
		printf("                <td>Team</td>\n");
	printf("            </tr>\n");
	nPlayers = atoi(si->numplayers);
	if(0 == nPlayers) {
		printf("            <tr>\n");
		printf("                <td colspan=\"4\">(none)</td>\n");
		printf("            </tr>\n");
	}
	else {
		for(i = 0; i < nPlayers; i++) {
			nTeam = *si->player[i].team == '1';
			printf("            <tr>\n");
			printf("                <td>%d</td>\n", i+1);
			printf("                <td class=\"%s\">%s</td>\n", nTeam ? "blue" : "red", si->player[i].name);
			printf("                <td>%s</td>\n", si->player[i].score);
			if('1' == *si->teamplay)
				printf("                <td class=\"%s\">%s</td>\n", nTeam ? "blue" : "red", nTeam ? "Blue" : "Red");
			printf("            </tr>\n");
		}
	}
	printf("        </table>\n");
}


void PrintFooter() {
	printf("        <hr/>\n");
	printf("        <div class=\"infotext\">Generated with Halo Query %s by ", VERSION_STRING);
	printf("<a href=\"mailto:brian@fieldeffect.net\">Brian Hurley</a>, <a href=\"http://www.divo.net/\">Andrew Chen</a></div>\n");
	printf("    </div>\n");
	printf("</body>\n");
	printf("</html>\n");
}


char *QueryServer(SOCKET fd, SERVER_INFO *si, unsigned char *status, int *statusSize) {
    struct sockaddr_in addr;
    int addrLen, err;

	unsigned char query[] = HALO_QUERY;

	/* set time to listen for a response */
#ifdef WIN32
	int timeout = TIMEOUT_VALUE * 1000; /* In milliseconds */
#else
	struct timeval timeout;
	memset(&timeout, 0, sizeof(timeout));
    timeout.tv_sec = TIMEOUT_VALUE; /* In seconds */
#endif

    addr.sin_family = AF_INET;
    addr.sin_port = htons((unsigned short)atoi(si->port));
	addr.sin_addr.s_addr = LookupServer(si->addr);

	/* send query packet */
	err = sendto(fd, (const char *)query, sizeof(query), 0, (struct sockaddr *)&addr, sizeof(addr));

	if(SOCKET_ERROR == err)
        return("The socket was unable to successfully send data to the server.");

	err = setsockopt(fd, SOL_SOCKET, SO_RCVTIMEO, (void *)&timeout, sizeof(timeout));

	if(SOCKET_ERROR == err)
        return("The socket was unable to successfuly set the timeout value.");

	/* receive server status */
	addrLen = sizeof(addr);
	err = recvfrom(fd, (char *)status, *statusSize, 0, (struct sockaddr *)&addr, &addrLen);
	*statusSize = err; 

    if(SOCKET_ERROR == err) {
		if(ETIMEDOUT == SockLastError())
			return("A timeout occurred while waiting to receive data from the server.");
		else
			return("The socket was unable to successfuly receive data from the server.");
    }

	return("");
}
